//********************************************************************
// ConsoleAppNPC.cpp
//********************************************************************
#include "ConsoleAppNPC.h"
#include "EbIntArr.h"

//--------------------------------------------------------------------
// Auxiliary globals used only in this module
//--------------------------------------------------------------------

LONG _par1;
LONG _par2;

//--------------------------------------------------------------------
// Supported pair properties
//--------------------------------------------------------------------
// Note: the functions
// DiffersByOneBinaryBit, IsCoprime, IsPowerOf, IsProperPowerOf,
// IsProperPower and IsPrime are part of the separate library
// EbStaticLib. Their declarations are in EbIntArr.h

BOOL NoCondition(LONG,LONG,LONG){return true;}

BOOL IsGrayEx(LONG i,LONG j,LONG param)
        {return DiffersByOneBinaryBit(i+param,j+param);}

BOOL IsCoprimeEx(LONG i,LONG j,LONG param)
        {return IsCoprime(i+param,j+param);}
BOOL IsNotCoprimeEx(LONG i,LONG j,LONG delta)
        {return !IsCoprime(i+delta,j+delta);}

BOOL IsDifBaseToN(LONG i,LONG j,LONG)
        {DWORD pe;if (!IsPowerOf(abs(i-j),_par1,&pe)) return false;
         if ((LONG)pe < _par2) return false; else return true;}
BOOL IsDifProperPowerOf(LONG i,LONG j,LONG)  // No solutions
        {return IsProperPowerOf(abs(i-j),_par1);}
BOOL IsProperPower(LONG i,LONG j,LONG)
        {return IsProperPower(abs(i-j));}

BOOL MinDif(LONG i,LONG j,LONG)
        {return (abs(i-j) >= _par1);}
BOOL MaxDif(LONG i,LONG j,LONG)
        {return (abs(i-j) <= _par1);}
BOOL MinMaxDif(LONG i,LONG j,LONG)
        {return (abs(i-j) >= _par1) && (abs(i-j) <= _par2);}

BOOL IsDifDivisorOfSum(LONG i,LONG j,LONG param)
        {return (abs(i-j)<1)? false:((i+param+j+param)%abs(i-j)==0);}
BOOL IsDifProperDivisorOfSum(LONG i,LONG j,LONG param)
        {return ((abs(i-j)<1)||(abs(i-j)==(i+param+j+param)))? false:
                ((i+param+j+param)%abs(i-j) == 0);}
BOOL IsDifProperDivisorOfSumGT1(LONG i,LONG j,LONG param)
        {return ((abs(i-j)<2)||(abs(i-j)==(i+param+j+param)))? false:
                ((i+param+j+param)%abs(i-j) == 0);}

BOOL IsCoprimeDifSum(LONG i,LONG j,LONG param)
        {return IsCoprime(abs(i-j),i+param+j+param);}
BOOL IsNotCoprimeDifSum(LONG i,LONG j,LONG param)
        {return !IsCoprime(abs(i-j),i+param+j+param);}

BOOL IsDifPrime(LONG i,LONG j,LONG)
        {return IsPrime(abs(i-j));}
BOOL IsSumPrime(LONG i,LONG j,LONG param)
        {return IsPrime(i+param+j+param);}
BOOL IsPrimeSumDif(LONG i,LONG j,LONG param)
        {return IsPrime(i+param+j+param)&&IsPrime(abs(i-j));}

//--------------------------------------------------------------------
// This Console Application menu
//--------------------------------------------------------------------

CONSOLE_MENU_ITEM NpcMenu[] =
//--------------------------------------------------------------------
{
  {"0 Void pair-property (no condition)",NoCond},
  {"1 Extended Gray",GrayEx},
  {"2 Coprimes (no solutions for not_coprimes)",Coprime},
  {"3 Dif = base^k, for some k>=kmin",DifBaseToN},
  {"4 Dif = b^k, for some b>1 and k>1",DifIsAPower},
  {"5 Dif >= a minimum",MinDif},
  {"6 Dif <= a maximum",MaxDif},
  {"7 Dif is a prime",PrimeDif},
  {"8 Sum is a prime",PrimeSum},
  {"9 Sum and difference are both primes",PrimeSumDif},
  {"a Dif >= minimum and <= maximum",MinMaxDif},
  {"b Dif is divisor of sum",DifDivisorOfSum},
  {"c Dif is a proper divisor of sum",DifProperDivisorOfSum},
  {"d Dif is a proper divisor of sum, dif > 1",DifProperDivisorOfSumGT1},
  {"e Dif and sum are coprime",DifSumCoprime},
  {"f Dif and sum are not coprime",DifSumNotCoprime},
  {NULL,NULL}
};

RETERR RunNpcApp()
//--------------------------------------------------------------------
// Main entry to this App
{
  RETERR err;
  while (true) {
    WelcomeHeader(
           "Neighbor Property Cycles (NPC) Utility _ Build 13.");
    CPrint(" Pertinent article: DOI 10.3247/SL5Math14.002\n");
    CPrint(" Select a pair-property and follow instructions:\n");
    err = Menu(NpcMenu);
    aFile.Close();
    bFile.Close();
    if (err == 0xffffffff) break;
    if (!err) CLPrint("Done; hit any key for a new run ");
    else CLPrint("Error; hit any key for a new run ");
    _getch(); // wait for a character (in conio)
  }
  return err;
}

RETERR NpcMode(BOOL (*condition)(LONG,LONG,LONG))
//--------------------------------------------------------------------
// Given an cyclen>0, find the first NPC of length cyclen (if any)
{
  REAL  count;
  DWORD cyclen,mode,lmin,lmax;
  LONG  offset;

  OpenConsoleFilesAB(&aFile);
  
  mode = CGetLong("Specify mode "
       "[1...First_NPC(L), 2...All_NPC(L), 3...NPC(Lmin..Lmax)]:");
  if ((mode < 1)||(mode > 3)) {CLPrintL("Illegal input");return 1;}

  switch (mode) {
    case 1: {
      cyclen = CGetLong("Specify L:");
      if (cyclen < 1) {CLPrintL("Illegal input");return 1;}
      offset = CGetLong("Cycle offset:");
      if (offset < 0) {CLPrintL("Illegal input");return 1;}
      CPrintL("");
      count = NPC(cyclen,mode,offset,condition);
      if (count == 0) {
        sprintf(Buffer,"cyclen = %li : none",cyclen);
        aFile.WriteLine(Buffer);
        CPrintL(" ",Buffer);
      }
    } break;
    case 2: {
        cyclen = CGetLong("Specify L:");
        if (cyclen < 1) {CLPrintL("Illegal input");return 1;}
        offset = CGetLong("Cycle offset:");
        if (offset < 0) {CLPrintL("Illegal input");return 1;}
        CPrintL("");
        sprintf(Buffer,"NPC's of cycle length %li",cyclen);
        aFile.WriteLine(Buffer);
        NPC(cyclen,mode,offset,condition);
    } break;
    default: {
      mode = 3;
      lmin = CGetLong("Specify Lmin:");
      if (lmin < 1) {CLPrintL("Illegal input");return 1;}
      lmax = CGetLong("Specify Lmax:");
      if (lmax < lmin) {CLPrintL("Illegal input");return 1;}
      offset = CGetLong("Cycle offset:");
      if (offset < 0) {CLPrintL("Illegal input");return 1;}
      CPrintL("");
      for (cyclen=lmin; cyclen<=lmax; cyclen++) {
        count = NPC(cyclen,mode,offset,condition);
        sprintf(Buffer,"%4li %12.0f",cyclen,count);
        aFile.WriteLine(Buffer);
        CPrintL(" ",Buffer);
      }
    }
  }
  return 0;
}

//--------------------------------------------------------------------
// Display & output file dumping function
//--------------------------------------------------------------------

void DumpNPC
        (DWORD cyclen,DWORD* npc,REAL count,DWORD mode,LONG offset)
//--------------------------------------------------------------------
{
  LPSTR ptr = Buffer;
  DWORD k;
  ptr += sprintf(ptr,"cyclen = %3li; count = %10.0f: ",cyclen,count);
  for (k=0; k<cyclen; k++) {
    ptr += sprintf(ptr," %li",npc[k]+offset);
  }
  CPrintL(" ",Buffer);
  aFile.WriteLine(Buffer);
}

//--------------------------------------------------------------------
// Pair-property condition functions for NpcMode()
//--------------------------------------------------------------------

RETERR NoCond()
//--------------------------------------------------------------------
{
  return NpcMode(NoCondition);
}

RETERR GrayEx()
//--------------------------------------------------------------------
{
  return NpcMode(IsGrayEx);
}

RETERR DifBaseToN()
//--------------------------------------------------------------------
{
  _par1 = CGetLong("Specify the base:");
  if (_par1 < 2) {CLPrintL("Illegal input");return 1;}
  _par2 = CGetLong(" Specify kmin:");
  if (_par2 < 0) {CLPrintL("Illegal input");return 1;}
  return NpcMode(IsDifBaseToN);
}

RETERR DifProperPowerOfBase()                          // No solutions
//--------------------------------------------------------------------
{
  _par1 = CGetLong("Specify the base:");
  if (_par1 < 2) {CLPrintL("Illegal input");return 1;}
  return NpcMode(IsDifProperPowerOf);
}

RETERR DifIsAPower()
//--------------------------------------------------------------------
{
  return NpcMode(IsProperPower);
}

RETERR MinDif()
//--------------------------------------------------------------------
{
  _par1 = CGetLong("Specify minimum difference:");
  if (_par1 < 1) {CLPrintL("Illegal input");return 1;}
  return NpcMode(MinDif);
}

RETERR MaxDif()
//--------------------------------------------------------------------
{
  _par1 = CGetLong("Specify maximum difference:");
  if (_par1 < 1) {CLPrintL("Illegal input");return 1;}
  return NpcMode(MaxDif);
}

RETERR MinMaxDif()
//--------------------------------------------------------------------
{
  _par1 = CGetLong("Specify minimum difference:");
  if (_par1 < 1) {CLPrintL("Illegal input");return 1;}
  _par2 = CGetLong(" Specify maximum difference:");
  if (_par2 < _par1) {CLPrintL("Illegal input");return 1;}
  return NpcMode(MinMaxDif);
}

RETERR PrimeDif()
//--------------------------------------------------------------------
{
  return NpcMode(IsDifPrime);
}

RETERR PrimeSum()
//--------------------------------------------------------------------
{
  return NpcMode(IsSumPrime);
}

RETERR PrimeSumDif()
//--------------------------------------------------------------------
{
  return NpcMode(IsPrimeSumDif);
}

RETERR Coprime()
//--------------------------------------------------------------------
{
  return NpcMode(IsCoprimeEx);
}

RETERR DifDivisorOfSum()
//--------------------------------------------------------------------
{
  return NpcMode(IsDifDivisorOfSum);
}

RETERR DifProperDivisorOfSum()
//--------------------------------------------------------------------
{
  return NpcMode(IsDifProperDivisorOfSum);
}

RETERR DifProperDivisorOfSumGT1()
//--------------------------------------------------------------------
{
  return NpcMode(IsDifProperDivisorOfSumGT1);
}

RETERR DifSumCoprime()
//--------------------------------------------------------------------
{
  return NpcMode(IsCoprimeDifSum);
}

RETERR DifSumNotCoprime()
//--------------------------------------------------------------------
{
  return NpcMode(IsNotCoprimeDifSum);
}


//********************************************************************
//********************************************************************
// Number crunching proper
//********************************************************************
//********************************************************************

REAL NPC(DWORD cyclen,DWORD mode,LONG offset,
         BOOL pairproperty(LONG i,LONG j,LONG offset))
//--------------------------------------------------------------------
// This is the arithmetic core of all NPC functions.
// mode can be 1, 2, or 3  and determines the following behavior:
// mode = 1: As soon as a NPC is found, it is dumped and
//           the function exits, returning 1.
// mode = 2: All found NPC's ar dumped, and
//           the function returns their count.
// mode = 3: All NPC's are  computed, but not dumped.
//           The function returns the NPC's count.
// If no NPC is found, nothing is dumped the returned value is 0.
// The argument BOOL pairproperty(LONG i,LONG j,LONG offset) points
// to the pair-property testing function for two neighbors with
// values i+offset and j+offset. It must return "true" when the pair
// satisfies the property.
// For the prototype of the "dumping" function DumpNPC,
// and an example of its implementation, see above.
// If modes 1 or 2 are used, DumpNPC must be supplied by the user,
// while in mode 3 DumpNPC is not used at all.
{
  REAL   count;           // DWORD count would be conceptually better,
    // but REAL can accomodate integers up to 52 bits without overflow

  DWORD  k,bktidx;                           // Backtracking variables
  DWORD  i,j,used,npc1max;                 // Element search variables

  count  = 0;                                        // Default result
  if (!cyclen) return count;                      // Zero cycle length

  DWORD* npc = new DWORD[cyclen];  // Allocate a new array for the npc
               // The stored npc values range from 0 through cyclen-1:
               // The "offset" is added only to carry out the tests
  npc[0] = 0;                    // First element in any npc must be 0

  if (cyclen > 2) {                                // When cyclen > 2:
                                        // Prepare optimization: start
    npc1max = cyclen;   // Find largest value compatible with 0+offset
    while (--npc1max) {
      if (npc1max && pairproperty(0,npc1max,offset)) break;
    }
    if (npc1max) {                       // And now the second largest
      while (--npc1max) {
        if (npc1max && pairproperty(0,npc1max,offset)) break;
      }
    }                                     // Prepare optimization: end
                                                   // Initializations:
    bktidx = 0;                          // Start backtrack index at 0
    k = 0;                                          // Dummy backtrack
    j = npc[k];              // Note: 0 will never be used again for j
    while (++k < cyclen) {   // npc[0] through npc[k] were done, go on
          // Now k was increased, and we are looking for a good npc[k]

      while (++j < cyclen) {    // Use j to scan all npc[k] candidates
            // starting just beyond where we left it in last backtrack
                                                // Optimization: start
        if ((k == 1) && (j > npc1max)) {j = cyclen; break;}
                                                  // Optimization: end
        used = 0;          // Make sure it was not yet used in the npc
        for (i=1;i<k;i++){if (npc[i]==j) {used=1;break;}}
        if (!used) {                               // If not yet used:
          // Test pair property between j and the last valid npc term:
          if (pairproperty(npc[k-1],j,offset)) {             // If ok:
            npc[k] = j;         // We have a good npc[k] candidate ...
            if ((k+1)<cyclen) {   // If this k is not the last one ...
              bktidx = k;     // ... just remember it for backtracking
            } else {          // But if this is the last k-term, then:
              if (j >= npc[1]) {   // test if the npc is canonical ...
                //...and do an extra pair-property test for cyclicity:
                if (pairproperty(npc[0],j,offset)) {
                  count = count+1;       // Found a good npc! Dump it?
                  if (mode<3) DumpNPC(cyclen,npc,count,mode,offset);
                }
              }
              j = cyclen;      // Last term forces j-scan termination!
            }
            break;
          }
        }
      }
      // We are done with this k. There are three possible situations:
      // k< cyclen-1 (an intermediate term) and j<cyclen (success)
      // k< cyclen-1 but no good j exists as indicated by j==cyclen
      // k==cyclen-1 (last term) and j was forcibly set to cyclen.

      if ((mode==1) && count) break;        // mode 1; first npc found

                              // Test whether we are done with this k?
      if (j < cyclen) {       // No, just reset the j-cycle for next k
        j = 0;
      } else {                                                 // Yes:
        if (!bktidx) break;          // Do we have pending backtracks?
        k = bktidx-1;                      // If yes, backtrack k, ...
        j = npc[bktidx];                      // ... initialize j, ...
        bktidx--;           // ... and decrement backtrack depth index
      }
    }
  } else if (cyclen==2) {              // Special case: cycle length 2
    npc[1] = 1;
        // Do the single test of the pair property for (npc[0],npc[1])
    if (pairproperty(0,1,offset)) {  // If satisfied, it counts as npc
      count = count+1;           // and can be dumped, if so specified
      if (mode < 3) DumpNPC(cyclen,npc,count,mode,offset);
    }
  } else if (cyclen==1) {              // Special case: cycle length 1
        // Do the single test of the pair property for (npc[0],npc[0])
    if (pairproperty(0,0,offset)) {  // If satisfied, it counts as npc
      count = count+1;           // and can be dumped, if so specified
      if (mode < 3) DumpNPC(cyclen,npc,count,mode,offset);
    }
  }
  delete [] npc;                    // Disallocate the temporary array
  return count;                               // and return the result
}

