// *******************************************************************
// File EbCpuAsm.h
// *******************************************************************
// Copyright @2014 by Stan Sykora and Extra Byte
#ifndef _EbCpuAsm_h_
#define _EbCpuAsm_h_

// Attention: This module uses WORD-aligned data

#include "EbInt64.h"

//--------------------------------------------------------------------
// EFLAG bits:
//--------------------------------------------------------------------
#define EFLAG_CF	0x00000001	// Carry
#define EFLAG_PF	0x00000004	// Parity
#define EFLAG_ACF	0x00000010	// Auxiliary Carry
#define EFLAG_ZF	0x00000040	// Zero
#define EFLAG_SF	0x00000080	// Sign
#define EFLAG_TF	0x00000100	// Trap
#define EFLAG_IF	0x00000200	// Interrupt enable
#define EFLAG_DF	0x00000400	// Direction
#define EFLAG_OF	0x00000800	// Overflow
#define EFLAG_IOPL	0x00003000	// I/O privilege level
#define EFLAG_NT	0x00004000	// Nested task
#define EFLAG_RF	0x00010000	// Resume
#define EFLAG_VM	0x00020000	// Virtual 8086 mode
#define EFLAG_AC	0x00040000	// Alignment Checks
#define EFLAG_VIF	0x00080000	// Virtual interrupt
#define EFLAG_VIP	0x00100000	// Virtual interrupt pending
#define EFLAG_ID	0x00200000	// CPUID instruction

//--------------------------------------------------------------------
// Edx1 features of CPUID:
//--------------------------------------------------------------------
#define FPU_FLAG        0x00000001	// FPU available
#define VME_FLAG        0x00000002	// Virtual Mode Extension
#define DE_FLAG         0x00000004	// Debugging Extension
#define PSE_FLAG        0x00000008	// Page-size Extensions
#define TSC_FLAG        0x00000010	// Time Stamp Counter
#define MSR_FLAG        0x00000020	// Model Specific Registers
#define PAE_FLAG        0x00000040	// Physical Address Extension
#define MCE_FLAG        0x00000080	// Machine Check Exceptions
#define CX8_FLAG        0x00000100	// CMPXCHG8B instruction
#define APIC_FLAG       0x00000200	// On-Chip APIC

#define SEP_FLAG        0x00000800	// SysEnter,SysExit instructions
#define MTRR_FLAG       0x00001000	// Memory Type Range Registers
#define PGE_FLAG        0x00002000	// Page Global Enable
#define MCA_FLAG        0x00004000	// Machine Check Architecture
#define CMOV_FLAG       0x00008000	// Conditional Move instruction
#define PAT_FLAG        0x00010000	// Page Attribute Table
#define PSE36_FLAG	0x00020000	// 36bit Page Size Extension
#define PSNUM_FLAG      0x00040000	// Processor serial number
#define CLFSH_FLAG      0x00080000	// CLFLUSH instruction

#define DS_FLAG		0x00200000	// Debug Store
#define ACPI_FLAG	0x00400000	// Thermal Monitor software
#define MMX_FLAG        0x00800000	// MMX instructions
#define FXSR_FLAG       0x01000000	// Fast FP Save&Restore
#define SSE_FLAG        0x02000000	// Streaming SIMD Extensions
#define SSE2_FLAG       0x04000000	// SSE2 Extensions
#define SS_FLAG		0x08000000	// Self Snoop
#define TM_FLAG         0x20000000	// Thermal Monitor controller

//--------------------------------------------------------------------
// ExEdx1 features of CPUID:
//--------------------------------------------------------------------
#define Ex_SSEMMX_FLAG	0x00400000	// SSE-MMX Extensions
#define Ex_3DNOW_FLAG	0x80000000	// 3DNOW (AMD only)
#define Ex_3DNOWEX_FLAG	0x40000000	// 3DNOW Extensions (AMD only)

//--------------------------------------------------------------------
// Machine setup structure
//--------------------------------------------------------------------

typedef struct {
  WORD  GDTLIMIT;		   // Global Descriptor Table Register
  DWORD GDTBASE;
  WORD  IDTLIMIT;		// Interrupt Descriptor Table Register
  DWORD IDTBASE;
  WORD  MSW;		    			// Machine status word
  WORD  LDT;			    // Local Descriptor Table Register
} MACHINE_TASK_SETUP;

//--------------------------------------------------------------------
// Machine specific register (MSR) info structure
//--------------------------------------------------------------------

typedef struct {
  DWORD RegisterAddress;
  DWORD LowDW;
  DWORD HighDW;
} MSR_INFO;

//--------------------------------------------------------------------
// Assembler functions
//--------------------------------------------------------------------
extern "C" {
                                               // Any Privilege Level:
  void  CpuIdInstruction(DWORD* eax,DWORD* ebx,DWORD* ecx,DWORD* edx);
  void  ReadMachineTaskSetup(MACHINE_TASK_SETUP* ms);
  void  EnableAlignementExceptions();
  void  DisableAlignementExceptions();
  void  ReadTSC(void*);
  DWORD AlignementExceptions();
  DWORD BsfCycles(DWORD iterations);

  DWORD CpuEFLAGS();
  DWORD CpuCF();
  DWORD CpuPF();
  DWORD CpuACF();
  DWORD CpuZF();
  DWORD CpuSF();
  DWORD CpuDF();
  DWORD CpuOF();

  DWORD GetESP();
  DWORD GetCS();
                             // Any privilege level under Windows95/98
        // Under NT4/2000/XP needs a driver to unlock the IO addresses
  void OutPortDirect(WORD port,BYTE data);
  BYTE InPortDirect(WORD port);
                                 // Ring0 only in all Operatin Systems
  void ReadMSR(MSR_INFO* msr);
  void WriteMSR(MSR_INFO* msr);
}

//--------------------------------------------------------------------
#endif


